"""
Change log:
5/14/2020 added ProgressBar class for display progress
5/14/2020 change revcomp to 4X faster
5/12/2020 added alignment module to file.
5/6/2020 added RNA structure and Vienna RNA warpping.
"""
__version__="0.1.12"